import { __awaiter } from "tslib";
import { AmplifyAppSyncSimulatorAuthenticationType as AuthTypes } from 'amplify-appsync-simulator';
import { invokeResource } from 'amplify-java-function-runtime-provider/lib/utils/invoke';
import { invoke } from 'amplify-nodejs-function-runtime-provider/lib/utils/invoke';
import axios from 'axios';
import fs from 'fs';
import { first, forEach, isNil } from 'lodash';
import { mergeTypes } from 'merge-graphql-schemas';
import path from 'path';
// @ts-ignore
import directLambdaRequest from './templates/direct-lambda.request.vtl';
// @ts-ignore
import directLambdaResponse from './templates/direct-lambda.response.vtl';
const directLambdaMappingTemplates = {
    request: directLambdaRequest,
    response: directLambdaResponse,
};
export const appSyncConfig = (context, appSyncConfig) => {
    // Flattening params
    const cfg = Object.assign(Object.assign({}, appSyncConfig), { mappingTemplates: (appSyncConfig.mappingTemplates || []).flat(), functionConfigurations: (appSyncConfig.functionConfigurations || []).flat(), dataSources: (appSyncConfig.dataSources || []).flat() });
    const mappingTemplatesLocation = path.join(context.serverless.config.servicePath, cfg.mappingTemplatesLocation || 'mapping-templates');
    const { defaultMappingTemplates = {} } = cfg;
    const getMappingTemplate = (filePath) => {
        return fs.readFileSync(path.join(mappingTemplatesLocation, filePath), {
            encoding: 'utf8',
        });
    };
    const getFileMap = (basePath, filePath) => ({
        path: filePath,
        content: fs.readFileSync(path.join(basePath, filePath), {
            encoding: 'utf8',
        }),
    });
    const makeDataSource = (source) => {
        if (source.name === undefined || source.type === undefined) {
            return null;
        }
        const dataSource = {
            name: source.name,
            type: source.type,
        };
        switch (source.type) {
            case 'AMAZON_DYNAMODB': {
                return Object.assign(Object.assign({}, dataSource), { config: Object.assign(Object.assign({}, context.options.dynamoDb), { tableName: source.config.tableName }) });
            }
            case 'AWS_LAMBDA': {
                const { functionName } = source.config;
                if (functionName === undefined) {
                    context.plugin.log(`${source.name} does not have a functionName`, {
                        color: 'orange',
                    });
                    return null;
                }
                const conf = context.options;
                if (conf.functions && conf.functions[functionName] !== undefined) {
                    const func = conf.functions[functionName];
                    return Object.assign(Object.assign({}, dataSource), { invoke: (payload) => __awaiter(void 0, void 0, void 0, function* () {
                            const result = yield axios.request({
                                url: func.url,
                                method: func.method,
                                data: payload,
                                validateStatus: () => false,
                            });
                            return result.data;
                        }) });
                }
                const func = context.serverless.service.functions[functionName];
                if (func === undefined) {
                    context.plugin.log(`The ${functionName} function is not defined`, {
                        color: 'orange',
                    });
                    return null;
                }
                if (func.runtime === 'java8' || func.runtime === 'java11') {
                    // @ts-ignore
                    return Object.assign(Object.assign({}, dataSource), { invoke: (payload) => invokeResource({
                            srcRoot: path.join(context.serverless.config.servicePath, context.options.location),
                            handler: func.handler,
                            // @ts-ignore
                            package: func.environment.servicePackage,
                            event: JSON.stringify(payload),
                            environment: Object.assign(Object.assign(Object.assign({}, (context.options.lambda.loadLocalEnv ? process.env : {})), context.serverless.service.provider.environment), func.environment),
                        }, context) });
                }
                else {
                    return Object.assign(Object.assign({}, dataSource), { invoke: (payload) => invoke({
                            packageFolder: path.join(context.serverless.config.servicePath, context.options.location),
                            handler: func.handler,
                            event: JSON.stringify(payload),
                            environment: Object.assign(Object.assign(Object.assign({}, (context.options.lambda.loadLocalEnv === true ? process.env : {})), context.serverless.service.provider.environment), func.environment),
                        }) });
                }
            }
            case 'AMAZON_ELASTICSEARCH':
            case 'HTTP': {
                return Object.assign(Object.assign({}, dataSource), { endpoint: source.config.endpoint });
            }
            default:
                return dataSource;
        }
    };
    const makeMappingTemplate = (resolver, type) => {
        const { name, type: parent, field, substitutions = {} } = resolver;
        const defaultTemplatePrefix = name || `${parent}.${field}`;
        const templatePath = !isNil(resolver === null || resolver === void 0 ? void 0 : resolver[type])
            ? resolver === null || resolver === void 0 ? void 0 : resolver[type]
            : !isNil(defaultMappingTemplates === null || defaultMappingTemplates === void 0 ? void 0 : defaultMappingTemplates[type])
                ? defaultMappingTemplates === null || defaultMappingTemplates === void 0 ? void 0 : defaultMappingTemplates[type]
                : `${defaultTemplatePrefix}.${type}.vtl`;
        let mappingTemplate;
        // Direct lambda
        // For direct lambdas, we use a default mapping template
        // See https://amzn.to/3ncV3Dz
        if (templatePath === false) {
            mappingTemplate = directLambdaMappingTemplates[type];
        }
        else {
            mappingTemplate = getMappingTemplate(templatePath);
            // Substitutions
            const allSubstitutions = Object.assign(Object.assign({}, cfg.substitutions), substitutions);
            forEach(allSubstitutions, (value, variable) => {
                const regExp = new RegExp(`\\$\{?${variable}}?`, 'g');
                mappingTemplate = mappingTemplate.replace(regExp, value);
            });
        }
        return mappingTemplate;
    };
    const makeResolver = (resolver) => {
        return {
            kind: resolver.kind || 'UNIT',
            fieldName: resolver.field,
            typeName: resolver.type,
            dataSourceName: resolver.dataSource,
            functions: resolver.functions,
            requestMappingTemplate: makeMappingTemplate(resolver, 'request'),
            responseMappingTemplate: makeMappingTemplate(resolver, 'response'),
        };
    };
    const makeFunctionConfiguration = (config) => ({
        dataSourceName: config.dataSource,
        name: config.name,
        requestMappingTemplate: makeMappingTemplate(config, 'request'),
        responseMappingTemplate: makeMappingTemplate(config, 'response'),
    });
    const makeAuthType = (authType) => {
        const auth = {
            authenticationType: authType.authenticationType,
        };
        if (auth.authenticationType === AuthTypes.AMAZON_COGNITO_USER_POOLS) {
            auth.cognitoUserPoolConfig = {
                AppIdClientRegex: authType.userPoolConfig.appIdClientRegex,
            };
        }
        else if (auth.authenticationType === AuthTypes.OPENID_CONNECT) {
            auth.openIDConnectConfig = {
                Issuer: authType.openIdConnectConfig.issuer,
                ClientId: authType.openIdConnectConfig.clientId,
            };
        }
        return auth;
    };
    const makeAppSync = (config) => ({
        name: config.name,
        apiKey: context.options.apiKey,
        defaultAuthenticationType: makeAuthType(config),
        additionalAuthenticationProviders: (config.additionalAuthenticationProviders || []).map(makeAuthType),
    });
    // Load the schema. If multiple provided, merge them
    const schemaPaths = Array.isArray(cfg.schema) ? cfg.schema : [cfg.schema || 'schema.graphql'];
    const schemas = schemaPaths.map((schemaPath) => getFileMap(context.serverless.config.servicePath, schemaPath));
    const schema = {
        // @ts-ignore
        path: first(schemas).path,
        content: mergeTypes(schemas.map((s) => s.content)),
    };
    return {
        appSync: makeAppSync(cfg),
        schema,
        resolvers: cfg.mappingTemplates.map(makeResolver),
        dataSources: cfg.dataSources.map(makeDataSource).filter((v) => v !== null),
        functions: cfg.functionConfigurations.map(makeFunctionConfiguration),
    };
};
